
rm(list=ls())

source("Functions/LoadRequiredPackages.R")
source("Functions/GeneralPreproces.R")
source("Functions/CraftAndImputeTrain.R")
source("Functions/CraftAndImputePredict.R")
source("Functions/QuitarOutliers.R")

source("Functions/GetXGBTreeModel.R")

packageList <- c("xgboost", "stringi","imager", "futile.logger", "cluster" ,"doParallel", "caret",
                 "lubridate", "magrittr", "purrr", "tidyverse")
LoadRequiredPackages(packageList)

dataModelarOrig <- read.delim("Data/Modelar_UH2019.txt",sep = "|" ,encoding = "UTF-8", na.strings = "")
dataPredictOrig <- read.delim("Data/Estimar_UH2019.txt",sep = "|" ,encoding = "UTF-8")
dataPredictOrig$TARGET <- 0
dataModelar <- GeneralPreproces(dataModelarOrig)
dataPredict <- GeneralPreproces(rbind(dataModelarOrig,dataPredictOrig)) %>% filter(HY_id %in% dataPredictOrig$HY_id)
#Vemos los NA
naByColumn <- dataModelar %>% map_dbl(~sum(is.na(.))) 
naByColumn[naByColumn %>% order(decreasing = T)]

#Observamos que desaparecen al craftear las variables
dataModelar <- CraftAndImputeTrain(dataModelar)
naByColumn <- dataModelar %>% map_dbl(~sum(is.na(.))) 
naByColumn[naByColumn %>% order(decreasing = T)]
# Quitamos outliers, aunque hicimos muchas graficas hasta llegar a la conclusion de lo que era o no era outlier.
dataModelar <- QuitarOutliers(dataModelar)

#### PRIMERAS GRAFICAS DEDENSIDAD
# Hay variables que si parecen provocar cambions en la distribucion del tiempo medio de visita en el inmueble.
ggplot(dataModelar, aes(TARGET)) +
    geom_density()+
    labs(title="Densidad del TARGET")
ggsave(paste0("Densidad_TARGET.png"), width = 12, height = 10)

#Realizamos Summary para entender la variable TARGET
summary(dataModelar$TARGET)
sd(dataModelar$TARGET)

ggplot(dataModelar %>% mutate(Tiene_Descripcion=ifelse(Tiene_Descripcion==0,"No tiene","Si tiene")), aes(TARGET)) +
    geom_density() +
    facet_wrap(~Tiene_Descripcion)+
    labs(title="Densidad del TARGET segun si el anuncio Tiene Descripcion")
ggsave(paste0("Densidad_TARGET_TieneDescripcion.png"), width = 12, height = 10)

ggplot(dataModelar %>% mutate(Tiene_Distribucion=ifelse(Tiene_Distribucion==0,"No tiene","Si tiene")), aes(TARGET)) +
    geom_density() +
    facet_wrap(~Tiene_Distribucion)+
    labs(title="Densidad del TARGET segun si el anuncio Tiene Distribucion")
ggsave(paste0("Densidad_TARGET_TieneDistribucion.png"), width = 12, height = 10)

ggplot(dataModelar %>% mutate(Tiene_Fotos=ifelse(Tiene_Fotos==0,"No tiene","Si tiene")), aes(TARGET)) +
    geom_density() +
    facet_wrap(~Tiene_Fotos)+
    labs(title="Densidad del TARGET segun si el anuncio Tiene foto")
ggsave(paste0("Densidad_TARGET_TieneFoto.png"), width = 12, height = 10)

ggplot(dataModelar, aes(TARGET)) +
    geom_density() +
    facet_wrap(~Exit_discreta)+
    labs(title="Densidad del TARGET segun el percentil de Tasa de Rebote")
ggsave(paste0("Densidad_TARGET_QuantileBounce.png"), width = 12, height = 10)

# En otras variables como Tiene_PrecioAnterior o Tiene_Antiguedad no vems demasiada distincion
# teniendo en cuenta que son pocas las que carecen de estos datos. 
ggplot(dataModelar %>% mutate(Tiene_Certificado=ifelse(Tiene_Certificado==0,"No tiene","Si tiene")), aes(TARGET)) +
    geom_density() +
    facet_wrap(~Tiene_PrecioAnterior)+
    labs(title="Densidad del TARGET segun si el anuncio Tiene certificado")
ggsave(paste0("Densidad_TARGET_TieneCertificado.png"), width = 12, height = 10)

ggplot(dataModelar %>% mutate(Tiene_Antiguedad=ifelse(Tiene_Antiguedad==0,"No tiene","Si tiene")), aes(TARGET)) +
    geom_density() +
    facet_wrap(~Tiene_Antiguedad)+
    labs(title="Densidad del TARGET segun si el anuncio Tiene certificado")
ggsave(paste0("Densidad_TARGET_TieneAntiguedad.png"), width = 12, height = 10)



# Explicacion del primer Cluster creado solamente con el objetivo de reducir provincias,
# teniendo en cuenta su PIB y densidad de poblacion.
ggplot(dataModelar %>% select(PIB_Capita,DensidadPob,Cluster,HY_provincia) %>% unique, aes(DensidadPob, PIB_Capita, col=Cluster)) +
    geom_point() +
    geom_text(aes(label=HY_provincia), size=3, nudge_y = 0.2)+
    labs(title="Cluster1 teniendo en cuenta solo el PIB y Densidad de la provincia")
ggsave(paste0("Explicacion_Clster1_SoloProvincias.png"), width = 12, height = 10)


# Visualizamos relaciones mas complejas, intentando graficar hasta 4 dimensiones en eje x,y,color y facet.
ggplot(dataModelar, aes(Tiene_Fotos,TARGET)) +
    geom_point()+
    labs(title="TARGET segun si el anuncio Tiene foto")
ggsave(paste0("TARGET_TieneFoto.png"), width = 12, height = 10)

ggplot(dataModelar, aes(Num_Fotos,TARGET)) +
    geom_point()+
    labs(title="TARGET segun Numero de Fotos")
ggsave(paste0("TARGET_NumFotos.png"), width = 12, height = 10)

ggplot(dataModelar, aes(Num_Fotos,TARGET)) +
    geom_point() + 
    facet_wrap(~HY_tipo)+
    labs(title="TARGET segun Numero de Fotos por Tipo")
ggsave(paste0("TARGET_NumFotos__Tipo.png"), width = 12, height = 10)

ggplot(dataModelar, aes(Num_Fotos,TARGET,col=HY_tipo)) +
    geom_point() + 
    facet_wrap(~HY_provincia)+
    labs(title="TARGET segun Numero de Fotos por Provincia & coloreado por Tipo")
ggsave(paste0("TARGET_NumFotos__Provincia__Tipo.png"), width = 12, height = 10)

ggplot(dataModelar, aes(Num_Fotos,TARGET,col=HY_tipo)) +
    geom_point() + 
    facet_wrap(~Cluster_TipoProvGA) +
    labs(title="TARGET segun Numero de Fotos dividido por Cluster2 & coloreado por Tipo")
ggsave(paste0("TARGET_Num_Fotos__Cluster__Tipo.png"), width = 12, height = 10)

ggplot(dataModelar, aes(Descript_Ncaracteres,TARGET,col=HY_tipo)) +
    geom_point() + 
    labs(title="TARGET segun Numero de caracteres de la descripcion coloreado por Tipo")
ggsave(paste0("TARGET_NumCaracteresDescripcion__Tipo.png"), width = 12, height = 10)

ggplot(dataModelar, aes(Distrib_Ncaracteres,TARGET,col=HY_tipo)) +
    geom_point() + 
    labs(title="TARGET segun Numero de caracteres de la distribucion coloreado por Tipo")
ggsave(paste0("TARGET_NumCaracteresDistribucion__Tipo.png"), width = 12, height = 10)

ggplot(dataModelar, aes(HY_antiguedad,TARGET,col=Cluster)) +
    geom_point() + 
    labs(title="TARGET segun Antiguedad coloreado por Cluster1 solo de provincias")
ggsave(paste0("TARGET_Antiguedad__ClusterProv.png"), width = 12, height = 10)

ggplot(dataModelar, aes(HY_precio,TARGET,col=Cluster)) +
    geom_point() + 
    facet_wrap(~HY_tipo) +
    labs(title="TARGET segun Antiguedad coloreado por Cluster1 solo de provincias")
ggsave(paste0("TARGET_Antiguedad__ClusterProv.png"), width = 12, height = 10)



# El peso relativo de cada Tiene Fotos (sí, no) es 75%-25%
dataModelar %>% count(Tiene_Fotos == 0)
dataModelar %>% count(Tiene_Descripcion == 0)
dataModelar %>% count(Tiene_Certificado == 0)
dataModelar %>% count(Tiene_PrecioAnterior == 0)


#Y obtenemos las densidades para ver la distribución
dataModelar %>% filter(Tiene_Fotos==0) %>% filter(GA_page_views>1)  %>% pull(TARGET) %>% densityplot()
dataModelar %>% filter(Tiene_Fotos==1) %>%filter(GA_page_views>1)   %>% pull(TARGET )%>% densityplot()


#Creamos las variables preciomedioprov, preciomediotipo y preciomedioprovytipo para ver si obtenemos más información sobre un precio segregado que nos ayude con el target
dataModelar %<>% left_join(
    dataModelar %>% group_by(HY_provincia) %>% summarise(Precio_Medio_Prov = mean(Precio_m2)),
    by="HY_provincia"
    
)
dataModelar %<>% left_join(
    dataModelar %>% group_by(HY_tipo) %>% summarise(Precio_Medio_Tipo = mean(Precio_m2)),
    by="HY_tipo"
    
)
dataModelar %<>% left_join(
    dataModelar %>% group_by(HY_provincia,HY_tipo) %>% summarise(Precio_Medio_ProvTipo = mean(Precio_m2))
    
)


# Probamos diferentes normalizaciones del precio por m2 para tratar de eliminar diferencias 
# entre el IPC por comunidades

dataModelar$Precio_m2_Norm <- dataModelar$Precio_m2 /dataModelar$Precio_Medio_Tipo
dataModelar$Precio_m2_Norm <- dataModelar$Precio_m2 /dataModelar$Precio_Medio_Prov
dataModelar$Precio_m2_Norm <- dataModelar$Precio_m2 /dataModelar$Precio_Medio_ProvTipo
dataModelar$Precio_m2_Norm <- dataModelar$Precio_m2 /( 0.5*dataModelar$Precio_Medio_Prov +0.5*dataModelar$Precio_Medio_Tipo)

ggplot(dataModelar, aes(TARGET))+
    geom_histogram(binwidth = 10)

ggplot(dataModelar, aes(TARGET))+  #Se observa una distribucion de cola larga en la derecha
    geom_density()

ggplot(dataModelar, aes(TARGET,HY_precio)) + 
    geom_point()

ggplot(dataModelar %>% filter(TARGET<500), aes(TARGET,Precio_m2)) + 
    geom_point() 

ggplot(dataModelar %>% filter(TARGET<500), aes(Precio_m2_Norm,TARGET, col=Tiene_Fotos)) + 
    geom_point()

ggplot(dataModelar %>% filter(GA_page_views<150),aes(GA_page_views,TARGET,col=N_Quincenas) ) +
    geom_point()


ggplot(dataModelar%>% filter(TARGET<200) %>% filter(GA_page_views<100),aes(GA_page_views,N_Quincenas,col=TARGET)) +
    geom_point()

#Vemos como el target permanece constante a partir de 300 visitas aprox
ggplot(dataModelar %>% filter(GA_page_views>400),aes(GA_page_views,TARGET)) +
    geom_point()



#Analizamos las correlaciones
corrPrecios<- round(cor(dataModelar[,c(13,43,50,51,61,62,72,73)]), 2)

corrplot:: corrplot.mixed(corrPrecios)

corrGA <- round(cor(dataModelar[,c(38:43,52, 64:67)]), 2)

corrplot:: corrplot.mixed(corrGA)

#Y las vemos con funciones de densidad

PerformanceAnalytics::chart.Correlation(dataModelar %>% select(HY_precio, Precio_m2, Views_interaccion, N_Quincenas,
                                                               VisitasInteraccionXquincena, DiffPerc_Precio,TARGET))

PerformanceAnalytics::chart.Correlation(dataModelar %>% select(GA_page_views, GA_exit_rate, VisitasXquincena,TARGET))


PerformanceAnalytics::chart.Correlation(dataModelar %>% select(TARGET, Precio_m2_Norm))


#Analizamos los boxplots
dataBoxplots <- dataModelar

#Observamos que las variables con menos visitas tienden a tener una media por visita más alta,
#la cual se va disipando conforme aumenta el número de visitas
boxplot(dataModelar$TARGET~dataModelar$Visitas_discreta, xlab= 'Percentiles Visitas',ylab = 'TARGET')

#Se observan las influencias de los outliers en las variables
boxplot(dataModelar[,c(39:40)])
boxplot(dataModelar$Precio_m2, main = 'Precio por m2')

#Cluster 2dimensional PIB/Densidad

datatip <- scale(unique(dataModelar[,c(52,53)]))

k1 <- kmeans(datatip, 4)

# plotcluster(unique(datatip), k1$cluster,)

datatip %<>% cbind(Cluster=as.factor(k1$cluster), Provincia = as.character(unique(dataModelar$HY_provincia))) %>% as_tibble()
datatip %<>% mutate(DensidadPob = as.numeric(DensidadPob),
                    PIB_Capita = as.numeric(PIB_Capita)) 
ggplot(datatip, aes(DensidadPob, PIB_Capita, col=as.factor(Cluster))) +
    geom_point() +
    geom_text(aes(label=Provincia), size=3, nudge_y = 0.2)


#Hemos realizado un análisis cluster para dividir las observaciones por provincia en cuatro grupos y analizar su PIB y su densidad de población

#Cluster 2

datatip2 <- scale(dataModelar[,c(52:53, 69:72)])

k2 <- kmeans(datatip2, 4)


datatip2 %<>% cbind(Cluster=as.numeric(k2$cluster), 
                    Provincia=dataModelar$HY_provincia,
                    TARGET=dataModelar$TARGET,
                    Precio_m2_Norm=dataModelar$Precio_m2_Norm) %>% as_tibble()

ggplot(datatip2, aes(Precio_m2_Norm, TARGET , colour=Provincia)) +
    geom_point() +
    geom_text(aes(label=Provincia), size=3, nudge_y = 0.2)+
    facet_wrap(~Cluster)

ggplot(datatip2%>% filter(TARGET < 180), aes(Precio_m2_Norm, TARGET , size=TARGET)) +
    geom_point() +
    geom_text(aes(label=Provincia), size=3, nudge_y = 0.2)+
    facet_wrap(~Cluster)

datatip2 %>% group_by(Cluster) %>% summarise(TargetMean = median(TARGET), 
                                             Count = n(), Sd = sd(TARGET))
datatip2 %>% filter(TARGET < 60*3) %>% 
    group_by(Cluster) %>% summarise(TargetMean = median(TARGET), 
                                             Count = n(), Sd = sd(TARGET))


#Se realiza otro análisis con el fin de observar grupos diferenciados en el TARGET y como influye un precio normalizado en éste
#Cluster para TARGET/Visitas_interaccion


datatip3 <- scale(unique(dataModelar[,c(43,63,70)]))

k3 <- kmeans(datatip3, 4)

# plotcluster(unique(datatip), k1$cluster,)

datatip3 %<>% cbind(Cluster=as.factor(k3$cluster)) %>% as_tibble()

ggplot(datatip3, aes(TARGET, Views_interaccion, col=as.factor(Cluster))) +
    geom_point() +
    geom_text(aes(label=TARGET), size=3, nudge_y = 0.2)
