CraftAndImputeTrain <- function(x, as_tibble=F, group ="HY_cod_postal"){
    tipos <- c('Almacen','Aparcamiento', 'Edificio turestico',
               'Otro', 'Edificio comercial', 'Trastero', 'Garaje', 'Solar', 'Suelo rustico no urbanizable')
    data <- x
    data %<>% mutate(HY_cert_energ = ifelse(is.na(HY_cert_energ) & HY_tipo %in% tipos, "NoAplica", 
                                      ifelse(is.na(HY_cert_energ),"DeberiaPeroNo",HY_cert_energ))) 
    data$HY_cert_energ <- data$HY_cert_energ %>% as.factor()
    
    #Eliminamos variables innecesarias con muchos NA
    
    data %<>% select(-IDEA_ind_liquidez)
    data %<>% select(-HY_metros_utiles)
    data %<>% select(-IDEA_pc_1960, -IDEA_pc_1960_69, -IDEA_pc_1970_79, -IDEA_pc_1980_89, 
                     -IDEA_pc_1990_99, -IDEA_pc_2000_10)

    
    #Imputamos NA de antigüedad
    
    data$HY_antiguedad[data$HY_antiguedad>2018]  <- 2008 ###Fuera outlier
    data %<>% left_join(
        data %>% select_(group,"HY_antiguedad") %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(AntiguedadGroupedMedian = median(HY_antiguedad)), by=group
    ) 
    
    # IMPUTACIONES POR MEDIANAS y MEDIANAS AGRUPADAS POR TIPO Y "GRUPO"
    data$AntiguedadGroupedMedian[is.na(data$AntiguedadGroupedMedian)] <- median(data$HY_antiguedad,na.rm = T)
    data %<>% mutate(HY_antiguedad = if_else(is.na(HY_antiguedad), AntiguedadGroupedMedian,HY_antiguedad),
                     HY_antiguedad = as.integer(HY_antiguedad)) 
    data %<>% left_join(
        data %>% select_(group, "IDEA_rent_alquiler") %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(RentAlqGroupedMedian = median(IDEA_rent_alquiler)), by=group
    ) 
    data$RentAlqGroupedMedian[is.na(data$RentAlqGroupedMedian)] <- median(data$IDEA_rent_alquiler,na.rm = T)
    data %<>% mutate(IDEA_rent_alquiler = if_else(is.na(IDEA_rent_alquiler), RentAlqGroupedMedian,IDEA_rent_alquiler),
                     IDEA_rent_alquiler = as.integer(IDEA_rent_alquiler))
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_unitprice_rent_residential" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(URResGroupedMedian = median(IDEA_unitprice_rent_residential )), by=group
    ) 
    data$URResGroupedMedian[is.na(data$URResGroupedMedian)] <- median(data$IDEA_unitprice_rent_residential ,na.rm = T)
    data %<>% mutate(IDEA_unitprice_rent_residential  = if_else(is.na(IDEA_unitprice_rent_residential ), URResGroupedMedian,IDEA_unitprice_rent_residential ),
                     IDEA_unitprice_rent_residential  = as.integer(IDEA_unitprice_rent_residential ))

        data %<>% left_join(
        data %>% select_(group, "IDEA_price_rent_residential" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(PRResGroupedMedian = median(IDEA_price_rent_residential )), by=group
    ) 
    data$PRResGroupedMedian[is.na(data$PRResGroupedMedian)] <- median(data$IDEA_price_rent_residential ,na.rm = T)
    data %<>% mutate(IDEA_price_rent_residential  = if_else(is.na(IDEA_price_rent_residential ), PRResGroupedMedian,IDEA_price_rent_residential ),
                     IDEA_price_rent_residential  = as.integer(IDEA_price_rent_residential ))
    
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_stock_rent_residential" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(SResGroupedMedian = median(IDEA_stock_rent_residential )), by=group
    ) 
    data$SResGroupedMedian[is.na(data$SResGroupedMedian)] <- median(data$IDEA_stock_rent_residential ,na.rm = T)
    data %<>% mutate(IDEA_stock_rent_residential  = ifelse(is.na(IDEA_stock_rent_residential ), SResGroupedMedian,IDEA_stock_rent_residential ),
                     IDEA_stock_rent_residential  = as.integer(IDEA_stock_rent_residential ))
    data$IDEA_stock_rent_residential <- data$IDEA_stock_rent_residential %>% as.integer()
    data %<>% left_join(
        data %>% select_(group, "IDEA_ind_tienda" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(ITiendaGroupedMedian = median(IDEA_ind_tienda )), by=group
    ) 
    data$ITiendaGroupedMedian[is.na(data$ITiendaGroupedMedian)] <- median(data$IDEA_ind_tienda ,na.rm = T)
    data %<>% mutate(IDEA_ind_tienda  = if_else(is.na(IDEA_ind_tienda ), ITiendaGroupedMedian,IDEA_ind_tienda ),
                     IDEA_ind_tienda  = as.integer(IDEA_ind_tienda ))
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_ind_turismo" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(ITurisGroupedMedian = median(IDEA_ind_turismo )), by=group
    ) 
    data$ITurisGroupedMedian[is.na(data$ITurisGroupedMedian)] <- median(data$IDEA_ind_turismo ,na.rm = T)
    data %<>% mutate(IDEA_ind_turismo  = if_else(is.na(IDEA_ind_turismo ), ITurisGroupedMedian,IDEA_ind_turismo ),
                     IDEA_ind_turismo  = as.integer(IDEA_ind_turismo ))
    
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_ind_alimentacion" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(IAliGroupedMedian = median(IDEA_ind_alimentacion )), by=group
    ) 
    data$IAliGroupedMedian[is.na(data$IAliGroupedMedian)] <- median(data$IDEA_ind_alimentacion ,na.rm = T)
    data %<>% mutate(IDEA_ind_alimentacion  = if_else(is.na(IDEA_ind_alimentacion ), IAliGroupedMedian,IDEA_ind_alimentacion ),
                     IDEA_ind_alimentacion  = as.integer(IDEA_ind_alimentacion ))
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_pc_comercio" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(IComGroupedMedian = median(IDEA_pc_comercio )), by=group
    ) 
    data$IComGroupedMedian[is.na(data$IComGroupedMedian)] <- median(data$IDEA_pc_comercio ,na.rm = T)
    data %<>% mutate(IDEA_pc_comercio  = if_else(is.na(IDEA_pc_comercio ), IComGroupedMedian,IDEA_pc_comercio ),
                     IDEA_pc_comercio  = as.integer(IDEA_pc_comercio ))
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_pc_industria" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(IIndusGroupedMedian = median(IDEA_pc_industria )), by=group
    ) 
    data$IIndusGroupedMedian[is.na(data$IIndusGroupedMedian)] <- median(data$IDEA_pc_industria ,na.rm = T)
    data %<>% mutate(IDEA_pc_industria  = if_else(is.na(IDEA_pc_industria ), IIndusGroupedMedian,IDEA_pc_industria ),
                     IDEA_pc_industria  = as.integer(IDEA_pc_industria ))
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_pc_oficina" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(PcOfiGroupedMedian = median(IDEA_pc_oficina )), by=group
    ) 
    data$PcOfiGroupedMedian[is.na(data$PcOfiGroupedMedian)] <- median(data$IDEA_pc_oficina ,na.rm = T)
    data %<>% mutate(IDEA_pc_oficina  = if_else(is.na(IDEA_pc_oficina ), PcOfiGroupedMedian,IDEA_pc_oficina ),
                     IDEA_pc_oficina  = as.integer(IDEA_pc_oficina ))
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_pc_otros" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(PcOtrosGroupedMedian = median(IDEA_pc_otros )), by=group
    ) 
    data$PcOtrosGroupedMedian[is.na(data$PcOtrosGroupedMedian)] <- median(data$IDEA_pc_otros ,na.rm = T)
    data %<>% mutate(IDEA_pc_otros  = if_else(is.na(IDEA_pc_otros ), PcOtrosGroupedMedian,IDEA_pc_otros ),
                     IDEA_pc_otros  = as.integer(IDEA_pc_otros ))
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_pc_residencial" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(PcResiGroupedMedian = median(IDEA_pc_residencial )), by=group
    ) 
    data$PcResiGroupedMedian[is.na(data$PcResiGroupedMedian)] <- median(data$IDEA_pc_residencial ,na.rm = T)
    data %<>% mutate(IDEA_pc_residencial  = if_else(is.na(IDEA_pc_residencial ), PcResiGroupedMedian,IDEA_pc_residencial ),
                     IDEA_pc_residencial  = as.integer(IDEA_pc_residencial ))
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_pc_trast_parking" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(PcParkGroupedMedian = median(IDEA_pc_trast_parking )), by=group
    ) 
    data$PcParkGroupedMedian[is.na(data$PcParkGroupedMedian)] <- median(data$IDEA_pc_trast_parking ,na.rm = T)
    data %<>% mutate(IDEA_pc_trast_parking  = if_else(is.na(IDEA_pc_trast_parking ), PcParkGroupedMedian,IDEA_pc_trast_parking ),
                     IDEA_pc_trast_parking  = as.integer(IDEA_pc_trast_parking ))
    
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_poblacion" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(PoblacionGroupedMedian = median(IDEA_poblacion )), by=group
    ) 
    data$PoblacionGroupedMedian[is.na(data$PoblacionGroupedMedian)] <- median(data$IDEA_poblacion ,na.rm = T)
    data %<>% mutate(IDEA_poblacion  = ifelse(is.na(IDEA_poblacion ), PoblacionGroupedMedian,IDEA_poblacion ),
                     IDEA_poblacion  = as.integer(IDEA_poblacion ))
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_densidad" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(DensidadGroupedMedian = median(IDEA_densidad )), by=group
    ) 
    data$DensidadGroupedMedian[is.na(data$DensidadGroupedMedian)] <- median(data$IDEA_densidad ,na.rm = T)
    data %<>% mutate(IDEA_densidad  = if_else(is.na(IDEA_densidad ), DensidadGroupedMedian,IDEA_densidad ),
                     IDEA_densidad  = as.integer(IDEA_densidad ))
    
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_unitprice_sale_residential" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(USResiGroupedMedian = median(IDEA_unitprice_sale_residential )), by=group
    ) 
    data$USResiGroupedMedian[is.na(data$USResiGroupedMedian)] <- median(data$IDEA_unitprice_sale_residential ,na.rm = T)
    data %<>% mutate(IDEA_unitprice_sale_residential  = if_else(is.na(IDEA_unitprice_sale_residential ), USResiGroupedMedian,IDEA_unitprice_sale_residential ),
                     IDEA_unitprice_sale_residential  = as.integer(IDEA_unitprice_sale_residential ))
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_price_sale_residential" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(PriSResiGroupedMedian = median(IDEA_price_sale_residential )), by=group
    ) 
    data$PriSResiGroupedMedian[is.na(data$PriSResiGroupedMedian)] <- median(data$IDEA_price_sale_residential ,na.rm = T)
    data %<>% mutate(IDEA_price_sale_residential  = if_else(is.na(IDEA_price_sale_residential ), PriSResiGroupedMedian,IDEA_price_sale_residential ),
                     IDEA_price_sale_residential  = as.integer(IDEA_price_sale_residential ))
    
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_stock_sale_residential" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(STSResiGroupedMedian = median(IDEA_stock_sale_residential )), by=group
    ) 
    data$STSResiGroupedMedian[is.na(data$STSResiGroupedMedian)] <- median(data$IDEA_stock_sale_residential ,na.rm = T)
    data %<>% mutate(IDEA_stock_sale_residential  = ifelse(is.na(IDEA_stock_sale_residential ), STSResiGroupedMedian,IDEA_stock_sale_residential ),
                     IDEA_stock_sale_residential  = as.integer(IDEA_stock_sale_residential ))
    
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_demand_sale_residential" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(DSResiGroupedMedian = median(IDEA_demand_sale_residential )), by=group
    ) 
    data$DSResiGroupedMedian[is.na(data$DSResiGroupedMedian)] <- median(data$IDEA_demand_sale_residential ,na.rm = T)
    data %<>% mutate(IDEA_demand_sale_residential  = if_else(is.na(IDEA_demand_sale_residential ), DSResiGroupedMedian,IDEA_demand_sale_residential ),
                     IDEA_demand_sale_residential  = as.integer(IDEA_demand_sale_residential ))
    
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_demand_rent_residential" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(DRResiGroupedMedian = median(IDEA_demand_rent_residential )), by=group
    ) 
    data$DRResiGroupedMedian[is.na(data$DRResiGroupedMedian)] <- median(data$IDEA_demand_rent_residential ,na.rm = T)
    data %<>% mutate(IDEA_demand_rent_residential  = if_else(is.na(IDEA_demand_rent_residential ), DRResiGroupedMedian,IDEA_demand_rent_residential ),
                     IDEA_demand_rent_residential  = as.integer(IDEA_demand_rent_residential ))
    
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_area" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(AreaGroupedMedian = median(IDEA_area )), by=group
    ) 
    data$AreaGroupedMedian[is.na(data$AreaGroupedMedian)] <- median(data$IDEA_area ,na.rm = T)
    data %<>% mutate(IDEA_area  = if_else(is.na(IDEA_area ), AreaGroupedMedian,IDEA_area ),
                     IDEA_area  = as.integer(IDEA_area ))
    
    
    data %<>% left_join(
        data %>% select_(group, "IDEA_ind_riqueza" ) %>% 
            na.omit() %>% group_by_(group)%>% 
            summarise(IndRiqGroupedMedian = median(IDEA_ind_riqueza )), by=group
    ) 
    data$IndRiqGroupedMedian[is.na(data$IndRiqGroupedMedian)] <- median(data$IDEA_ind_riqueza ,na.rm = T)
    data %<>% mutate(IDEA_ind_riqueza  = if_else(is.na(IDEA_ind_riqueza ), IndRiqGroupedMedian,IDEA_ind_riqueza ),
                     IDEA_ind_riqueza  = as.integer(IDEA_ind_riqueza ))
    
    data %<>% select(-contains("GroupedMedian")) 
    
    
    data$IDEA_ind_elasticidad[is.na(data$IDEA_ind_elasticidad)] <- median(data$IDEA_ind_elasticidad, na.rm = T)
    

    # Agrupamos por NUEVOS CLUSTERS

    data %<>% left_join(
    data %>% group_by(Cluster_TipoProvPreciom2) %>% summarise(Cluster_TPPm2_Target = median(TARGET))
    )
    data %<>% left_join(
        data %>% group_by(Cluster_TipoProvGA) %>% summarise(Cluster_TipoProvGA_Target = median(TARGET))
    )
    
    data %<>% left_join(
        data %>% group_by(Cluster_TipoProvPreciom2) %>% summarise(Cluster_TPPm2_GA_page_views = median(GA_page_views))
    )
    data %<>% left_join(
        data %>% group_by(Cluster_TipoProvGA) %>% summarise(Cluster_TipoProvGA_GA_page_views = median(GA_page_views))
    )
    
    data %<>% left_join(
        data %>% group_by(Cluster_TipoProvPreciom2) %>% summarise(Cluster_TPPm2_Precio_m2 = median(Precio_m2))
    )
    data %<>% left_join(
        data %>% group_by(Cluster_TipoProvGA) %>% summarise(Cluster_TipoProvGA_Precio_m2 = median(Precio_m2))
    )
    
    data %<>% left_join(
        data %>% group_by(Cluster_TipoProvPreciom2) %>% summarise(Cluster_TPPm2_Precio_m2 = median(Views_interaccion))
    )
    data %<>% left_join(
        data %>% group_by(Cluster_TipoProvGA) %>% summarise(Cluster_TipoProvGA_Precio_m2 = median(Views_interaccion))
    )
    
    data %<>% left_join(
        data %>% group_by(Cluster_TipoProvPreciom2) %>% summarise(Cluster_TPPm2_Num_Fotos = median(Num_Fotos))
    )
    data %<>% left_join(
        data %>% group_by(Cluster_TipoProvGA) %>% summarise(Cluster_TipoProvGA_Num_Fotos = median(Num_Fotos))
    )
    data %<>% left_join(
        data %>% group_by(Cluster_TipoProvPreciom2) %>% summarise(Cluster_TPPm2_VisitasXquincena = median(VisitasXquincena))
    )
    data %<>% left_join(
        data %>% group_by(Cluster_TipoProvGA) %>% summarise(Cluster_TipoProvGA_VisitasXquincena = median(VisitasXquincena))
    )
    
    
    dataCrafted <- data
    if(as_tibble)
        dataCrafted <- data %>% as_tibble()
    return(dataCrafted)
}

