GetXGBTreeModel <- function(trainSet_orig,nFolds = 4, mode = "CV", verbose = F, maximize = F, saveTrainPreds = T, 
                            plotImportances = F, customFolds = NULL, validationSet = NULL, print_every_n = 500, 
                            early_stopping_rounds = 100, nrounds = 3000, seed = 1234,logistic = F, obj = NULL,
                            # params = list(eta = .25, max_depth = 3, gamma = 2, alpha = 0.5, lambda = 1,
                            params = list(eta = .25, max_depth = 6, gamma = 1, alpha = 0.5, lambda = 1,
                                          # colsample_bytree = 0.9, min_child_weight = 4, subsample = .9)){
                                          colsample_bytree = 0.9, min_child_weight = 2, subsample = .9)){
                            # params = list(eta = .15, max_depth = 6, gamma = 0, alpha = 0.5, lambda = 1,
                            #               # colsample_bytree = 0.9, min_child_weight = 4, subsample = .9)){
                            #               colsample_bylevel = 0.67, min_child_weight = 1, subsample = .7)){
    
    
    MDAE <- function(m,f){
        median(abs(m-f))
    }
    feval <- function(preds, dtrain){
        labels <- getinfo(dtrain, "label")
        err <- MDAE(m = labels,f=preds)
        return(list(metric = "MDAE", value = err))
    }
    
    trainPreds <- NULL
    randomParams <- F
    if(tolower(params[1]) == "random"){
        randomParams <- T
        
        params <- expand.grid(eta = c(.2, .3,.4),
                              max_depth = c(3, 5,7),
                              gamma = c(0, 1, 2), 
                              alpha = c(0, 1),
                              lambda = c(.5, 1),
                              colsample_bytree = c(.7, .9),
                              min_child_weight = c(4, 6, 8, 12),
                              subsample = c(.85,0.95, 1)
        ) %>% 
            as_tibble() %>%
            sample_n(1)
        
        params <- list(eta = params$eta, 
                       max_depth = params$max_depth,
                       gamma = params$gamma,
                       alpha = params$alpha, 
                       lambda = params$lambda,
                       colsample_bytree = params$colsample_bytree,
                       min_child_weight = params$min_child_weight,
                       subsample = params$subsample)
    }
    
    
    switch(mode, 
           "CV" = {
               flog.info("Training XGBTree model with CV...")
               
               
               trainSet_xgb <- xgb.DMatrix(data.matrix(trainSet_orig[,-1]), label = data.matrix(trainSet_orig[,1]))
               
               
               if(!is.null(customFolds)){
                   set.seed(seed)
                   xgbCV <- xgb.cv(data = trainSet_xgb,
                                   verbose = verbose,
                                   nrounds = nrounds,
                                   seed = seed,
                                   folds = customFolds,
                                   feval = feval,
                                   maximize = maximize, 
                                   params = params,obj = obj,
                                   prediction = saveTrainPreds,
                                   print_every_n = print_every_n,
                                   early_stopping_rounds = early_stopping_rounds)
               } else {
                   set.seed(seed) 
                   xgbCV <- xgb.cv(data = trainSet_xgb,
                                   verbose = verbose,
                                   nrounds = nrounds,
                                   seed = seed,
                                   nfold = nFolds,
                                   feval = feval,
                                   maximize = maximize, 
                                   params = params,obj = obj,
                                   prediction = saveTrainPreds,
                                   print_every_n = print_every_n,
                                   early_stopping_rounds = early_stopping_rounds)
               }
               
               trainPreds <- xgbCV$pred
               bestIter <- xgbCV$best_iteration
               
               set.seed(seed)
               xgbModel <- xgb.train(data = trainSet_xgb,
                                     params = params,
                                     feval = feval,
                                     maximize = maximize,
                                     seed = seed,obj = obj,
                                     print_every_n = print_every_n,
                                     nrounds = bestIter)
               nRounds <- bestIter
           },
           "fullOptim" = {
               params = list(eta = .1, max_depth = 5, gamma = 0, #alpha = 0.5, lambda = 1,
                             colsample_bytree = 0.8, min_child_weight = 1, subsample = .8)
               
               
           }
           
    )
    
    imp_matrix <- NULL
    mostImportantFeature <- NULL
    
    tryCatch({
        imp_matrix <- xgb.importance(feature_names = colnames(trainSet_xgb), model = xgbModel)
        mostImportantFeature <-  imp_matrix$Feature %>% first() 

    },
    error=function(e){}
    )
    
    try(params$nRounds <- nRounds)
    
    xgb <- list(model = xgbModel, 
                trainPreds = trainPreds, 
                mostImportantFeature = mostImportantFeature, 
                importanceMatrix = imp_matrix,
                colNames = names(trainSet_orig),
                params = params)
    
    return(xgb)
}


