###############
## Catboost  ##
###############

vars <- c("TARGET", "HY_provincia", "HY_tipo", "HY_metros_totales", "HY_num_banos", "HY_num_terrazas",
          "HY_ascensor", "HY_trastero", "HY_num_garajes", "HY_precio","N_Quincenas", "DiffPerc_Precio", 
          "Precio_m2","Page_View_Per_Qincena","Cluster","HY_cert_energ")
    if(nModels==5) roundsTries <- c(6,7,7,9,7) else roundsTries <- runif(nModels,6,9) 
    
if(section == "train"){
    trainSetMethod <- foldMethodList$models_0$baseCraft[[nm]]
    trainSetMethod %<>% select(one_of(vars), contains("GA"), contains("Tiene")) 
    trainSetMethod %<>% na.omit() 
    
    # Despues de probar, quitamos los outliers antes de dividir entre train y validacion del catboost
    trainSetMethod <- QuitarOutliers(trainSetMethod,mode="filterOut",rounds=roundsTries[nm])
    flog.info("Roundtries: %s, custom fold %s",roundsTries[nm],cf)
    trainSetMethodNew <- trainSetMethod %>% sample_frac(1.6,replace=T)
    validationSetMethod <- setdiff(trainSetMethod,trainSetMethodNew)
    # trainSetMethodNew <- QuitarOutliers(trainSetMethodNew,mode="filterOut",rounds=roundsTries[nm])
    
    train_pool <- catboost.load_pool(data = trainSetMethodNew %>% select(-TARGET), label = trainSetMethodNew %>% pull(TARGET))
    validation_pool <- catboost.load_pool(data = validationSetMethod %>% select(-TARGET), label = validationSetMethod %>% pull(TARGET))
    model <- catboost.train(train_pool,  validation_pool,
                            params <- list(iterations=10000,
                                           learning_rate=0.2,
                                           depth=7,
                                           loss_function='MAE',
                                           eval_metric='MAE',
                                           random_seed = seed,
                                           od_type='Iter',
                                           metric_period = 500,
                                           od_wait=30,
                                           use_best_model=TRUE))
    return(model)
    
} else if(section == "predict"){
    testSetMethod <- foldPredsList$models_0$baseCraft[[nm]]
    testSetMethod <- testSetMethod %>% select(one_of(vars),contains("GA"), contains("Tiene"))
    model <- allCVMethodListBaseModels[[cf]]$models_1$catboost[[nm]]
    
    real_pool <- catboost.load_pool(testSetMethod[,-1])
    
    predictions <- catboost.predict(model, real_pool)
    
    return(predictions)
    
}  else if(section == "predict_ooSample"){
    testSetMethod <- ooSamplePredsList$models_0$baseCraft[[nm]]$test%>% arrange(HY_id)
    testSetMethod <- testSetMethod %>% select(one_of(vars), contains("GA"), contains("Tiene"))
    real_pool <- catboost.load_pool(testSetMethod[,-1])
    
    trainSetMethod <- ooSamplePredsList$models_0$baseCraft[[nm]]$train
    trainSetMethod <- trainSetMethod %>% select(one_of(vars), contains("GA"), contains("Tiene")) %>% na.omit()
    trainSetMethod <- QuitarOutliers(trainSetMethod,mode="filterOut",rounds=roundsTries[nm])
    
    trainSetMethodNew <- trainSetMethod %>% sample_frac(1.6,replace=T)
    validationSetMethod <- setdiff(trainSetMethod,trainSetMethodNew)
    train_pool <- catboost.load_pool(data = trainSetMethodNew %>% select(-TARGET), label = trainSetMethodNew %>% pull(TARGET))
    validation_pool <- catboost.load_pool(data = validationSetMethod %>% select(-TARGET), label = validationSetMethod %>% pull(TARGET))
    
    model <- catboost.train(train_pool,  validation_pool,
                            params <- list(iterations=10000,
                                           learning_rate=0.2,
                                           depth=7,
                                           loss_function='MAE',
                                           eval_metric='MAE',
                                           random_seed = seed,
                                           od_type='Iter',
                                           metric_period = 500,
                                           od_wait=30,
                                           use_best_model=TRUE))
    predictions <- catboost.predict(model, real_pool)
    
    # Capamos las predicciones que esten por debajo de 0
    return(pmax(predictions,0))
    
}

