# Recibe dataset y aplica principalmente 2 filtros. El primero relacionado con outliers univariantes
# y el segundo con iteraciones de una regresión lineal al TARGET y usando la distancia de Cooks sobre
# los valores fitados.

QuitarOutliers <- function(trainSet, mode="filterOut",rounds=3, toleranceDev=2, distOutlier=F, distOutlierCrit=10){
  trainSetSinOutliersInic <- trainSet %>% as_tibble()
  vars <- c("HY_precio","TARGET", 'GA_page_views','GA_mean_bounce', 'GA_mean_bounce',
            'Precio_m2', 'Visitas_T_C','VisitasXquincena', 'DiffPerc_Precio',
            'Views_interaccion', 'VisitasInteraccionXquincena')
  means <- trainSetSinOutliersInic %>% select(one_of(vars)) %>% summarise_all(mean)
  sds <- trainSetSinOutliersInic %>% select(one_of(vars)) %>% summarise_all(sd)
 
  # Opcionalmente permite usar el Z-core, distancia mahalanobis y boxplots o Tukey para detectar outliers
  #bajo un umbral de tolerancia distOutlierCrit.
  
  if(distOutlier){
      
      isnt_out_z <- function(x, thres = 3, na.rm = TRUE) {
          abs(x - mean(x, na.rm = na.rm)) <= thres * sd(x, na.rm = na.rm)
      }
      
      maha_dist <- . %>% select_if(is.numeric) %>%
          mahalanobis(center = colMeans(.), cov = cov(.))
      
      isnt_out_maha <- function(tbl, isnt_out_f, ...) {
          tbl %>% maha_dist() %>% isnt_out_f(...)
      }
      
      isnt_out_tukey <- function(x, k = 1.5, na.rm = TRUE) {
          quar <- quantile(x, probs = c(0.25, 0.75), na.rm = na.rm)
          iqr <- diff(quar)
          
          (quar[1] - k * iqr <= x) & (x <= quar[2] + k * iqr)
      }
      isnt_out_mad <- function(x, thres = 3, na.rm = TRUE) {
          abs(x - median(x, na.rm = na.rm)) <= thres * mad(x, na.rm = na.rm)
      }
      
      isnt_out_funs <- funs(
          z = isnt_out_z,
          # mad = isnt_out_mad,
          tukey = isnt_out_tukey
      )
      outMatrix <- trainSetSinOutliersInic %>% transmute_if(is.numeric, isnt_out_funs) %>% 
          mutate(IsOut = pmap_dbl(., lift_vd(sum))) 
      outMatrix %<>% cbind(Index= 1:nrow(outMatrix))
      outMatrix %<>% 
          filter(IsOut >= ncol(outMatrix)-distOutlierCrit)
      trainSetSinOutliersInic <- trainSetSinOutliersInic[outMatrix$Index,]
  }
  
  # TRATAMOS PRECIO Y TARGET DE MANERA ESPECIAL por su importancia.
  
  # Siempre quitamos los outliers de precio que esten mas de toleranceDev+1 veces la desviación del precio
  trainSetSinOutliersInic %<>% filter(between(HY_precio,means$HY_precio-(toleranceDev+1)*sds$HY_precio,
                        means$HY_precio+(toleranceDev+1)*sds$HY_precio))
  
  # Siempre tratamos los outliers en el TARGET que esten mas de toleranceDev veces la desviación del precio
  # O bien los quitamos o bien los capamos a la mediana
  if(mode=="filterOut"){
      trainSetSinOutliersInic %<>% filter(between(TARGET,means$TARGET-toleranceDev*sds$TARGET,
                        means$TARGET+toleranceDev*sds$TARGET))
      # #Precio_m2
      # trainSetSinOutliers %<>% filter(between(Precio_m2,means$Precio_m2-toleranceDev*sds$Precio_m2,
      #                                         means$Precio_m2+toleranceDev*sds$Precio_m2))

  } else if(mode=="cap"){
      
      trainSetSinOutliersInic$TARGET[trainSetSinOutliersInic$TARGET>means$TARGET+toleranceDev*sds$TARGET] <- means$TARGET+tolerance*sds$TARGET
  }

  trainSetSinOutliers <- trainSetSinOutliersInic
  # Repetimos rounds veces la regresion lineal al TARGET y la eliminacion de observaciones outliers
  for(n in 1:rounds){
      mod <- lm(TARGET ~ ., data=trainSetSinOutliers%>% select(TARGET,Cluster,HY_tipo,HY_cert_energ,contains("Tiene"),contains("Foto") ))
      cooksd <- cooks.distance(mod)
      influential <-cooksd[cooksd> 4*mean(cooksd, na.rm=T)] %>% names %>% na.omit() # influential row numbers
      outRows <- trainSetSinOutliers[influential, ]
      trainSetSinOutliers <- suppressMessages(anti_join(trainSetSinOutliers,outRows))
  }
  actualRows <- trainSetSinOutliers %>% nrow
  entryRows <- trainSet %>% nrow
if((entryRows-actualRows)>0.45*nrow(trainSet)){
    for(n in 1:2){
        mod <- lm(TARGET ~ ., data=trainSetSinOutliersInic%>% select(TARGET,Cluster,HY_tipo,HY_cert_energ,contains("Tiene"),contains("Foto") ))
        cooksd <- cooks.distance(mod)
        influential <-cooksd[cooksd> 4*mean(cooksd, na.rm=T)] %>% names %>% na.omit() # influential row numbers
        outRows <- trainSetSinOutliersInic[influential, ]
        trainSetSinOutliersInic <- suppressMessages(anti_join(trainSetSinOutliersInic,outRows))
    }
    cat( "Too many rows for filter or cap, returning the original DF.")
    flog.warn(cat( "Too many rows for filter or cap, returning the original DF."))
    return(trainSetSinOutliersInic)
}
  cat( (entryRows-actualRows)," rows have been removed or capped.")
  
  
  return(trainSetSinOutliers) 
}
