######################
#   Image Modeling   # 
######################

# Cargamos funciones y paquetes

source("Functions/LoadRequiredPackages.R")

packageList <- c("xgboost", "stringi", "magick","cluster","imager","EBImage","jpeg","keras", "imager",
                 "doParallel", "stringi", "caret","lubridate", "magrittr", "purrr",
                 "tidyverse")
LoadRequiredPackages(packageList)

imagesPath <- "./Data/imagenes_inmuebles_haya/"
filesTrain <- list.files(paste0(imagesPath,"train"),pattern = "jpg")
filesTest <- list.files(paste0(imagesPath,"test"),pattern = "jpg")


# Importamos el modelo preentenado para hacer Transfer Learning y predecir sobre nuestro dataset.
model <- application_vgg16(weights = "imagenet", 
                           include_top = FALSE)
model

image_prep <- function(x) {
  arrays <- lapply(x, function(path) {
    img <- image_load(path, target_size = c(224, 224))
    x <- image_to_array(img)
    x <- array_reshape(x, c(1, dim(x)))
    x <- imagenet_preprocess_input(x)
  })
  do.call(abind::abind, c(arrays, list(along = 1)))
}

# Cargamos nuestras imagenes
imagesPathTrain <- ".Data/imagenes_inmuebles_haya/train"
file_list_train <-  list.files(imagesPathTrain,pattern = "jpg", full.names = TRUE)
file_list_train <- file_list_train[1:700]
vgg16_feature_list <- data.frame()

# Predecimos cada imagen
for (image in file_list_train) {
  
  print(image)
  cat("Image", which(file_list_train == image), "from", length(file_list_train))
  vgg16_feature <- predict(model, image_prep(image))
  
  flatten <- as.data.frame.table(vgg16_feature, responseName = "value") %>%
    select(value)
  flatten <- cbind(image, as.data.frame(t(flatten)))
  
  vgg16_feature_list <- rbind(vgg16_feature_list, flatten)
}
# Como tarda mucho lo guardamos para poder cargarlo posteriormente
save(vgg16_feature_list, file = "./Data/imagenes_inmuebles_haya/train/feature/vgg16_feature_list.RData")

# Llevamos a cabo el PCA sobre las activaciones de la penultima capa del modelo preentrenado
pca <- prcomp(vgg16_feature_list[, -1],
              center = TRUE,
              scale = FALSE)
save(pca, file = "./Data/imagenes_inmuebles_haya/train/feature/pca.RData")

data.frame(PC1 = pca$x[, 1], 
           PC2 = pca$x[, 2]) %>%
  ggplot(aes(x = PC1, y = PC2)) +
  geom_point()

# Clusterizamo 10 componentes principales con el objetivo de que nuestras imagenes no labeleadas queden agrupadas y 
# se explicite la distancia entre ellas
set.seed(50)
cluster_pca <- kmeans(pca$x[, 1:10], 5)
cluster_feature <- kmeans(vgg16_feature_list[, -1], 5)

cluster_list <- data.frame(cluster_pca = cluster_pca$cluster, 
                           cluster_feature = cluster_feature$cluster,
                           vgg16_feature_list) %>%
  select(cluster_pca, cluster_feature, image) %>%
  mutate(class = gsub("./Data/imagenes_inmuebles_haya/train/feature/", "", image),
         class = substr(class, start = 1, stop = 3))
head(cluster_list)

cluster_list %>%
  count(cluster_pca, class)

cluster_list %>%
  count(cluster_feature, class)

cluster_list %>%
  mutate(PC1 = pca$x[, 1],
         PC2 = pca$x[, 2]) %>%
  ggplot(aes(x = PC1, y = PC2, color = class, shape = factor(cluster_pca))) +
  geom_point()

# Como son demasiadas imagenes, ploteamos imagenes aleatorias pertenecientes a cada cluster para comprobar si lo esta haciendo bien.
plot_random_images <- function(n_img = 4,
                               cluster = 1,
                               rows = 2,
                               cols = 2) {
  cluster_list_random_cl_1 <- cluster_list %>%
    filter(cluster_feature == cluster) %>%
    sample_n(n_img)
  
  graphics::layout(matrix(c(1:n_img), rows, cols, byrow = TRUE))
  for (i in 1:n_img) {
    path <- as.character(cluster_list_random_cl_1$image[i])
    img <- load.image(path)
    plot(img, axes = FALSE)
    title(main = paste("Cluster PCA", cluster))
  }
}


sapply(c(1:5), function(x) plot_random_images(cluster = x))


train_datagen <- image_data_generator(rescale = 1/255)
validation_datagen <- image_data_generator(rescale = 1/255)
train_generator <- flow_images_from_directory(
  paste0(imagesPath,"train"),
  train_datagen,
  target_size = c(150, 150),
  batch_size = 20,
  class_mode = "binary"
)
validation_generator <- flow_images_from_directory(
  validation_dir,
  validation_datagen,
  target_size = c(150, 150),
  batch_size = 20,
  class__mode = "binary"
)

batch <- generator_next(train_generator)
