######################
# Image Preprocesing #
######################

# Cargamos funciones y paquetes
source("Functions/LoadRequiredPackages.R")
source("Functions/FileRename.R")

packageList <- c("xgboost", "stringi","jpeg" ,"keras","cluster","imager", "futile.logger", "doParallel", "stringi", "caret",
                 "lubridate", "magrittr", "purrr", "tidyverse")
LoadRequiredPackages(packageList)

saveObject <- F      # Si queremos guardar los resultados. En origen lo pusimo a T

# Cargamos los datos
trainSet <- read.delim("Data/Modelar_UH2019.txt", sep = "|", encoding = "UTF-8", na.strings = "") 
predictSet <- read.delim("Data/Estimar_UH2019.txt", sep = "|", encoding = "UTF-8", na.strings = "") 


######################
# Sacamos los inmuebles que tienen foto. (No se adjuntam las imagenes, se deben poner en esta ruta)
imagesPath <- "./Data/imagenes_inmuebles_haya/"
files <- list.files(imagesPath,pattern = "jpg")

ids <- stri_extract_first_regex(files, "[0-9]+")
uniqueIds <- ids %>% unique

trainImageIds <- trainSet$HY_id[trainSet$HY_id%in%uniqueIds] 
predictImageIds <- predictSet$HY_id[predictSet$HY_id%in%uniqueIds]
all <- c(trainImageIds,predictImageIds)

if(saveObject) write_rds(all,"Data/ImageIds.RDS")


#################### 
# Sacamos el numero de fotos
imageCount <- tibble(Ids = ids, Count = 1)
imageCount %<>% 
  group_by(Ids) %>% 
  summarise(Num_Fotos = sum(Count))

if(saveObject) write_rds(imageCount, "Data/ImageIdsCount.RDS")


####################
# MODELAR - Sacamos las dimensiones y luminancia de las imagenes subidas 
#
# Acabamos un objeto con la mediana de dimensiones y brillo y luminosidad por anuncio, 
# otro con las dimensiones de todas las fotos y el histograma de brillo completo por anuncio
# y el ultimo que simplemente almacena las ids de anuncios con foto en el entrenamiento

trainImageResolution <- vector("list", length = length(trainImageIds)) %>% set_names(trainImageIds)
trainImageList <- vector("list", length = length(trainImageIds)) %>% set_names(trainImageIds)

fileDims <- data.frame()
fileDimsFull <- data.frame() 
luminImgDf <- NULL
for(id in trainImageIds %>% as.character()){
  idFiles <- files[stri_extract_first_regex(files, "[0-9]+")==id]
  fileDimsTemp <- data.frame() 
  fileDimsFullTemp <- data.frame() 
  for(file in 1:length(idFiles)){
    fileName <- idFiles[file]
    img <- load.image(paste0(imagesPath,fileName))
    luminImg <- grayscale(img) %>% hist %>% pluck("density")
    luminImgDf <- luminImgDf %>% rbind(luminImg)
    imageDim <- readJPEG(paste0(imagesPath,fileName)) %>%
      dim %>% as.matrix() %>% t %>% as.data.frame()
    if(ncol(imageDim)!=3) imageDim <- imageDim %>% cbind(V3=1)
    fileDimsTemp <- fileDimsTemp %>% rbind(imageDim %>%
                                             set_names(c("w","h","c"))%>% cbind(medianLum = median(luminImg)))  #height x width x channels
    if(file==1) fileDimsFullTemp <- fileDimsTemp[file,1:2]
    fileDimsFullTemp <-fileDimsFullTemp %>% cbind(fileDimsTemp[file,1:2] %>% set_names("w","h"))
    # Las cortamos de la raiz de imagenes a una carpeta esecifica de train
    FileRename(from = paste0(imagesPath,fileName),
               to = paste0(imagesPath,"train/",fileName))
  }
  fileDims <- fileDims %>% bind_rows(fileDimsTemp %>% summarise_all(median) %>% add_column(Id=id))
  fileDimsFull <- fileDimsFull %>% bind_rows(fileDimsFullTemp %>%  add_column(Id=id))
  trainImageList[[id]] %<>% append(idFiles)
  print(id)
}

if(saveObject){
  saveRDS(fileDims,"Data/FileDimsTrain.RDS")
  saveRDS(fileDimsFull,"Data/FileDimsFullTrain.RDS")
  saveRDS(trainImageResolution, "Data/trainImageList.RDS")
}

####################
# PESTIMAR- Sacamos las dimensiones y luminancia de las imagenes subidas
#
# Acabamos un objeto con la mediana de dimensiones y brillo y luminosidad por anuncio, 
# otro con las dimensiones de todas las fotos y el histograma de brillo completo por anuncio
# y el ultimo que simplemente almacena las ids de anuncios con foto en el predict

predictImageResolution <- vector("list", length = length(predictImageIds)) %>% set_names(predictImageIds)
predictImageList <- vector("list", length = length(predictImageIds)) %>% set_names(predictImageIds)

fileDims <- data.frame()
fileDimsFull <- data.frame() 
luminImgDf <- NULL
for(id in predictImageIds %>% as.character()){
  idFiles <- files[stri_extract_first_regex(files, "[0-9]+")==id]
  fileDimsTemp <- data.frame() 
  fileDimsFullTemp <- data.frame() 
  for(file in 1:length(idFiles)){
    fileName <- idFiles[file]
    img <- load.image(paste0(imagesPath,fileName))
    luminImg <- grayscale(img) %>% hist %>% pluck("density")
    luminImgDf <- luminImgDf %>% rbind(luminImg)
    imageDim <- readJPEG(paste0(imagesPath,fileName)) %>%
      dim %>% as.matrix() %>% t %>% as.data.frame()
    if(ncol(imageDim)!=3) imageDim <- imageDim %>% cbind(V3=1)
    fileDimsTemp <- fileDimsTemp %>% rbind(imageDim %>%
                                             set_names(c("w","h","c"))%>% cbind(medianLum = median(luminImg)))  #height x width x channels
    if(file==1) fileDimsFullTemp <- fileDimsTemp[file,1:2]
    fileDimsFullTemp <-fileDimsFullTemp %>% cbind(fileDimsTemp[file,1:2] %>% set_names("w","h"))
    # Las cortamos de la raiz de imagenes a una carpeta esecifica de predict
    FileRename(from = paste0(imagesPath,fileName),
               to = paste0(imagesPath,"predict/",fileName))
  }
  fileDims <- fileDims %>% bind_rows(fileDimsTemp %>% summarise_all(median) %>% add_column(Id=id))
  fileDimsFull <- fileDimsFull %>% bind_rows(fileDimsFullTemp %>%  add_column(Id=id))
  predictImageList[[id]] %<>% append(idFiles)
  print(id)
  
}

if(saveObject){
  saveRDS(fileDims,"Data/FileDimsPredict.RDS")
  saveRDS(fileDimsFull,"Data/FileDimsFullPredict.RDS")
  saveRDS(predictImageList, "Data/predictImageList.RDS")
}






















