## CajaMar UniversityHack Datathon 2019 ##
##       Equipo GoDeeperOrGoHome        ##
##########################################
# Estructura del codigo
#
# 1. Carga de funciones, paquetes y settings
# 2. Carga de datos y primer tratamiento
# 3. Bucle de cf Folds (iterar hasta el numero de folds escogido en settings)
### 3.1 Entrenamiento sobre folds-1 
### 3.2 Prediccion del fold sobrante
### 3.3 Errores en ese fold
# 4. Errores promedios de todos los folds
# 5. Entrenamiento y prediccion OutOfSample
# 6. Entrenamiento y prediccion sobre los datos a entregar


# Limpimos el global environment
rm(list=ls())

# Cargamos nuestras funciones propias
source("Functions/LoadRequiredPackages.R")
source("Functions/GeneralPreproces.R")
source("Functions/GetXGBTreeModel.R")
source("Functions/MethodSwitch.R")
source("Functions/QuitarOutliers.R")


# Cargamos los paquetes y si no los tenemos nuestra funcion nos los instala directamente
packageList <- c("xgboost", "keras","catboost", "stringi","arules","imager","fastDummies","ruler",
                 "futile.logger", "doParallel", "cluster","factoextra", "caret", "e1071", "robustbase",
                 "lubridate", "magrittr", "purrr", "tidyverse")
LoadRequiredPackages(packageList)
select <- dplyr::select
filter <- dplyr::filter

#### Settings
exportPreds <- T                    # Si se quiere predecir el dataset a enviar a Cajamar
seed <- 120420190
ooSampleLength <- 1000              # Numero de filas del conjunto OutOfSample
nFolds <- 3                         # Numero de folds en los que dividiremos los datos sin incluir la muestra el ooSample
testFoldPerc <- 1/nFolds            # Si quisiesemos un tamaño de cada fold de test determinado.
nModels <- 5                        # Las veces que se repite cada method o algoritmo, cambiando params, pesos iniciales y boostrapeando el train
nModelsReduceFunction <- median     # Funcion de agrupacion de los nModels por familia de algoritmo.

# Definimos los metodos o modelos por nivel
baseMethodsByLevel <- list(models_0 = c("baseCraft_OneHot", "base","baseCraft"),
                           models_1 = c("abcrf","catboost","catboost_newCraft","svm",
                                        "xgb_tunnedCustom_newCraft", "xgb_tunnedCustom"),
                           models_2 = c("stack_mean", "stack_median"))


if(!dir.exists("Logs")) dir.create("Logs")
flog.appender(appender.file(file = sprintf("Logs/%s_log.txt", format(today(), format="%Y%m%d"))))
flog.info("##########################################################")
flog.info("########## REAL STATE MODELING - Group: Go Deeper or Go Home ############")
flog.info("##########################################################")
flog.info("READING & CRAFTING THE DATA")

tryCatch({
  # Cargamos los datos y hacemos las transformaciones genericas (NAs y variables que no queremos que dependan
  # de train o test. Andaremos con ojo para no crear leackage de informacion entre sets y sobreestimacion del
  # accuracy de los modelos. Habrá un craft especifico en el train y test para el resto de variables, imputaciones...)
  
  data <- read.delim("Data/Modelar_UH2019.txt", sep = "|", encoding = "UTF-8", na.strings = "") 
  dataCrafted <- GeneralPreproces(data)
  
  # Deteccion de nas por variable
  # dataCrafted %>% map_dbl(~sum(is.na(.)))
},
error=function(e){
  errorMessage <<- sprintf("Error Loading or Crafting the data :  %s ",  e$message)
  flog.warn(errorMessage)
})


tryCatch({
  # Creamos la muestra OutOfSample y definimos como trainSetOriginal el resto.
  ooSampleData <- dataCrafted %>% sample_n(ooSampleLength)
  trainSetOriginal <- suppressMessages(anti_join(dataCrafted,ooSampleData))
  
  # Creamos los folds sobre trainSerOriginal
  foldIndex <- tibble(index = seq_len(nrow(trainSetOriginal)), 
                      fold = round(runif(nrow(trainSetOriginal),1,nFolds)))
  customFolds <- split(foldIndex$index, f = foldIndex$fold)
  trainSetSplited <-  vector("list", length = nFolds) %>% set_names(names(customFolds))
  
  # Inicializamos las listas donde guardaremos los outputs
  allCVMethodListBaseModels <- vector("list",nFolds) %>% set_names(1:nFolds)
  allCVPredsListBaseModels <- vector("list",nFolds) %>% set_names(1:nFolds)
  allCVPredsDfBaseModels <- vector("list",nFolds) %>% set_names(1:nFolds)
  allCVErrorsBaseModels <- vector("list",nFolds) %>% set_names(1:nFolds)
  megaTablePredsCV <- vector("list",nFolds) %>% set_names(1:nFolds)
  
  # Bucle principal de entrenamiento y predicción por fold
  for(cf in 1:nFolds){
    tryCatch({
      # Creamos lo que sera el train y el test de cada iteracion del CV  
      testSet <-  trainSetOriginal[customFolds[[cf]],] %>% arrange(HY_id)%>% arrange(HY_id)
      trainSet <- trainSetOriginal[setdiff(1:nrow(trainSetOriginal),customFolds[[cf]]),]
      
      ## TRAIN DEL cf FOLD CV
      # LLamamos al script de entrenamiento del que saldrán los modelos ya entrenados al environment
      flog.info(sprintf("Lets Train fold : %s", cf))
      source("Functions/Train.R")
      
      # Guardamos los modelos outputs del fold en su lista
      allCVMethodListBaseModels[[cf]] %<>% append(foldMethodList)
    },
    error=function(e){
      errorMessage <<- sprintf("Error in Training in Fold %s and Method %s:  %s ", cf, method,e$message)
    })
    tryCatch({
      flog.info(sprintf("Lets Predict fold : %s", cf))
      
      ## PREDICT DEL cf FOLD CV
      # Llamamos al script de prediccion 
      source("Functions/Predict.R")
      
      # Guardamos las preds outputs del fold en su lista
      allCVPredsListBaseModels[[cf]] %<>% append(foldPredsList)
      
      # Sacamos los errores del Fold y los mostramos en consola
      allCVPredsDfBaseModels[[cf]] <- foldPredsList[[2]]  %>%
        map_dfr(~ .x  %>% map(~ ifelse(.x < 0, 0, .x)), .id="Model") %>%
        select(Model, Observation, everything())
      predsBaseModels <- allCVPredsDfBaseModels[[cf]] %>% 
        mutate(FinalPred = pmap_dbl(allCVPredsDfBaseModels[[cf]]
                                    %>% select(-Model,-Observation), lift_vd(nModelsReduceFunction)))
      megaTablePredsCV[[cf]] <- left_join(predsBaseModels, testSet %>% select(HY_id, TARGET), by=c("Observation"="HY_id"))
      
      # Ploteamos la distribución de TARGET y las predicciones.
      # plotly::ggplotly(
      print(ggplot(megaTablePredsCV[[cf]], aes(FinalPred, col=Model)) + geom_density() + 
              geom_density(data=testSet %>% select(HY_id, TARGET) %>%
                             mutate(Model="TARGET"),aes(TARGET,col=Model))+
              labs(title=paste0("Fold: ",cf)))
      # )
      ggsave(paste0("Results/Fold",cf,".png"))
      print(megaTablePredsCV[[cf]] %>% mutate(Error=(FinalPred-TARGET)) %>% filter(Error<500) %>% pull(Error) %>% hist())
      
      ## ERRORES DEL cf FOLD CV
      med_ad <- megaTablePredsCV[[cf]] %>% 
        mutate(Error = abs(FinalPred-TARGET)) %>% group_by(Model) %>% summarise(MED_AD = median(Error))
      mse <- megaTablePredsCV[[cf]] %>% 
        mutate(Error = abs(FinalPred-TARGET)^2) %>% group_by(Model) %>% summarise(MSE = mean(Error))
      
      # Error de la mediana de las predicciones agregadas 
      medianFoldError <- megaTablePredsCV[[cf]] %>% group_by(Observation) %>%
        summarise(MeanPreds = median(FinalPred)) %>% 
        left_join(testSet %>% select(HY_id,TARGET),by=c("Observation" = "HY_id")) %>%
        mutate(Error = abs(MeanPreds-TARGET)) %>% summarise(MED_AD = median(Error)) %>% pull
      
      errorTableBaseModels <- inner_join(med_ad,mse, by="Model")
      allCVErrorsBaseModels[[cf]] <- errorTableBaseModels
      print(errorTableBaseModels)
      print(medianFoldError)
      
      # Para hacer testeos entre nModels
      megaTablePredsCV$`1`%>% filter(Model=="svm") %>% mutate_at(c("FinalPred",as.character(1:5)), funs(abs(.-TARGET)))  %>%
        summarise_all(median)
      
      # saveRDS(allCVMethodListBaseModels,"Data/allCVMethodListBaseModels.RDS")
      # saveRDS(allCVPredsDfBaseModels,"Data/allCVPredsDfBaseModels.RDS")
      # saveRDS(allCVPredsListBaseModels,"Data/allCVPredsListBaseModels.RDS")
      # saveRDS(ooSampleData,"Data/ooSampleData.RDS")
      
    },
    error=function(e){
      errorMessage <<- sprintf("Error in %s in Fold %s:  %s , method: %s ",section ,cf, method, e$message)
      flog.warn(errorMessage)
    })
    
  }
  
  ## ERRORES CONJUNTOS DEL CV
  # Sacamos las tablas de errores conjunta de todos los folds predichos de los baseModels
  finelErrorTable <- megaTablePredsCV %>% map_dfr(~.x %>% mutate(Error = abs(FinalPred - TARGET))) %>% group_by(Model) %>%
    summarise(MED_AD = median(Error), MSE = mean(Error^2), Sum = sum(Error))
  print("FINAL ERROR TABLE")
  print(finelErrorTable)
  write.csv(finelErrorTable,"Results/CVFinalTable.csv")
  
  ## ENTRENAMIENTO Y PREDICCION OUT OF SAMPLE
  # Predicciones Out Of Sample y sacamos la tabla de errores
  flog.info(sprintf("Starting OutOfSample step"))
  trainSet <- trainSetOriginal
  testSet <-  ooSampleData %>% arrange(HY_id)
  source("Functions/Predict_ooSample.R")
  
  # Sacamos las predicciones de cada algoritmo, condensando cada pasada de nModel de nivel 1 y 2
  ooSampleBaseModelDf <- left_join(ooSamplePredsList[[2]] %>%
                                     map_dfr(~ .x  %>% map(~ ifelse(.x < 0, 0, .x)), .id="Model") %>% select(Model, Observation, everything()) %>% 
                                     mutate(FinalPred = pmap_dbl(select(.,-Model,-Observation), lift_vd(nModelsReduceFunction))),
                                   testSet %>% select(HY_id, TARGET), by=c("Observation"="HY_id")) %>% 
    mutate(Error = abs(FinalPred - TARGET))
  ooSampleStackModelDf <- left_join(ooSamplePredsList[[3]] %>%
                                      map_dfr(~ .x  %>% map(~ ifelse(.x < 0, 0, .x)), .id="Model") %>% select(Model, Observation, everything()) %>% 
                                      mutate(FinalPred = pmap_dbl(select(.,-Model,-Observation), lift_vd(nModelsReduceFunction))),
                                    testSet %>% select(HY_id, TARGET), by=c("Observation"="HY_id")) %>% 
    mutate(Error = abs(FinalPred - TARGET)) 
  
  # Para analizar los diferentes nModels de un algoritmo
  ooSampleBaseModelDf %>% filter(Model=="catboost_newCraft") %>%  mutate_at(c("FinalPred",as.character(1:nModels)), funs(abs(.-TARGET)))  %>%
    summarise_all(median)
  
  # Tablas de Errores sobre el conjunto OOSample
  ooSampleBaseModelErrorTable <- ooSampleBaseModelDf %>% group_by(Model) %>%
    summarise(MED_AD = median(Error), MAE = mean(Error), MSE = mean(Error^2), RMSE = mean(Error^2)^(1/2), Sum = sum(Error), Sd = sd(Error), Max = max(Error), Min=min(Error) )
  ooSampleStackModelErrorTable <- ooSampleStackModelDf %>% group_by(Model) %>%
    summarise(MED_AD = median(Error), MAE = mean(Error), MSE = mean(Error^2), RMSE = mean(Error^2)^(1/2), Sum = sum(Error), Sd = sd(Error), Max = max(Error), Min=min(Error) )
  ooSampleErrorTable <-  ooSampleBaseModelErrorTable %>% bind_rows(ooSampleStackModelErrorTable)
  
  print("ooSAMPLE ERROR TABLE")
  print(ooSampleErrorTable %>% arrange(MED_AD))
  # browser()
  write.csv(ooSampleErrorTable %>% arrange(MED_AD),"Results/ooSampleFinalTable.csv")
  
  ## PREDICCIONES DEL DATASET A ENTREGAR
  if(exportPreds){
    # Crafteamos los datos de test y modelamos/predecimos con los modelos por nivel escogidos
    predData <- read.delim("Data/Estimar_UH2019.txt", sep = "|", encoding = "UTF-8", na.strings = "") 
    allDataCrafted <- GeneralPreproces(bind_rows(data,predData)) 
    predDataCrafted <- allDataCrafted %>% filter(HY_id %in% predData$HY_id) %>% arrange(HY_id)
    trainDataCrafted <- allDataCrafted %>% filter(HY_id %in% data$HY_id) %>% arrange(HY_id)
    trainSet <- trainDataCrafted
    testSet <- predDataCrafted
    testSet$TARGET <- 0
    
    # Entrenamiento y prediccion
    source("Functions/Predict_ooSample.R")
    
    # Condensacion de las predicciones de los difetentes modelos(methods) y diferentes lanzamientos(nModels)
    predictionsBaseDF <- ooSamplePredsList[[2]] %>%
      map_dfr(~ .x  %>% map(~ ifelse(.x < 0, 0, .x)), .id="Model") %>% select(Model, Observation, everything()) %>% 
      mutate(FinalPred = pmap_dbl(select(.,-Model,-Observation), lift_vd(nModelsReduceFunction)))
    predictionsStackModelDF <- ooSamplePredsList[[3]] %>%
      map_dfr(~ .x  %>% map(~ ifelse(.x < 0, 0, .x)), .id="Model") %>% select(Model, Observation, everything()) %>% 
      mutate(FinalPred = pmap_dbl(select(.,-Model,-Observation), lift_vd(nModelsReduceFunction))) 
    
    # Comparamos las predicciones sacadas con un sampleo del TARGET del trainSet
    TrainSampled <- dataCrafted %>% sample_n(1104) 
    
    plotly::ggplotly(ggplot(predictionsBaseDF, aes(FinalPred, col=Model)) + geom_density() + 
                       geom_density(data=TrainSampled %>% select(HY_id, TARGET) %>%
                                      mutate(Model="TARGET"),aes(TARGET,col=Model))+
                       labs(title=paste0("Sending Preds ")))
    ggsave(paste0("Results/Sending.png"))
    
    plotly::ggplotly(ggplot(predictionsStackModelDF, aes(FinalPred, col=Model)) + geom_density() + 
                       geom_density(data=TrainSampled %>% select(HY_id, TARGET) %>%
                                      mutate(Model="TARGET"),aes(TARGET,col=Model))+
                       labs(title=paste0("Sending Preds ")))
    ggsave(paste0("Results/SendingStack.png"))
    # browser()
    # Predicciones definitivas
    predictions <- predictionsStackModelDF %>% filter(Model=="stack_median") %>%
      select(Observation, FinalPred) %>% set_names(c("HY_id", "TM_Est"))
    
    write.table(x = predictions, file = "Data/preds_enviadas/aenviar.txt", 
                sep = "|",
                dec = ".", fileEncoding = "UTF-8", row.names = F)
    # Exportamos las predicciones definitivas
    write.table(x = predictions, file = "Results/GoDeeperOrGoHomeDef.txt", 
                sep = "|",
                dec = ".", fileEncoding = "UTF-8", row.names = F)
    
  }
  
},
error=function(e){
  if(!exists("section")) section <- "Before training"
  if(!exists("method")) method <- "Before methods"
  errorMessage <<- sprintf("FATAL ERROR IN SECTION %s, METHOD %s:  %s ", section, method,  e$message)
  print(errorMessage)
  flog.fatal(errorMessage)
})